Unit Parsing;

Interface

Function Parse (S : String; Arg : Real; Var Result : Real) : Byte;

(****************************************************)
(*                    Examples:                     *)
(* S = "4 * 60 + 10 * 9"         => Result = 330.0  *)
(* S = "1.5E2+2.3e-1"            => Result = 150.23 *)
(* S = "-(X-3)*(x+5)", Arg = 4.0 => Result = -9.0   *)
(* S = "9.0^0.5"                 => Result = 3.0    *)
(* S = "x%8", Arg = 15.0         => Result = 7      *)
(****************************************************)

(****************************************************)
(*                 Return values:                   *)
(* 0 - successfuly - no error                       *)
(* 1 - error in numeric format like "1e" or "2.e3"  *)
(* 2 - unknown symbol in expression                 *)
(* 3 - not balanced brakets i.e. "(x-1)*(x+2"       *)
(* 4 - positional error like in "4+x-" or ")2"      *)
(* 5 - expression is empty - nothing to compute     *)
(* 6 - math error such as division by zero etc.     *)
(****************************************************)

Implementation

Const
  StackSize = 128;
  SP1 : Word = 0;
  SP2 : Word = 0;
  Set1 : Array [0..8] of Char = #0'(+-*/%^)';
  Table1 : Array [0..7,0..8] of Byte = (
  { End (  +  -  *  /  %  ^  ) }
    (0, 1, 1, 1, 1, 1, 1, 1, 5), { Empty }
    (5, 1, 1, 1, 1, 1, 1, 1, 3), {   (   }
    (4, 1, 2, 2, 1, 1, 1, 1, 4), {   +   }
    (4, 1, 2, 2, 1, 1, 1, 1, 4), {   -   }
    (4, 1, 4, 4, 2, 2, 2, 1, 4), {   *   }
    (4, 1, 4, 4, 2, 2, 2, 1, 4), {   /   }
    (4, 1, 4, 4, 2, 2, 2, 1, 4), {   %   }
    (4, 1, 4, 4, 4, 4, 4, 2, 4)  {   ^   }
  );
  Table2 : Array [0..9,0..9] of Byte = (
  { End (  +  -  *  /  %  ^  )  Number }
    (0, 0, 0, 0, 1, 1, 1, 1, 1, 0), { Empty }
    (1, 0, 0, 0, 1, 1, 1, 1, 1, 0), {   (   }
    (1, 0, 1, 1, 1, 1, 1, 1, 1, 0), {   +   }
    (1, 0, 1, 1, 1, 1, 1, 1, 1, 0), {   -   }
    (1, 0, 1, 1, 1, 1, 1, 1, 1, 0), {   *   }
    (1, 0, 1, 1, 1, 1, 1, 1, 1, 0), {   /   }
    (1, 0, 1, 1, 1, 1, 1, 1, 1, 0), {   %   }
    (1, 0, 1, 1, 1, 1, 1, 1, 1, 0), {   ^   }
    (0, 1, 0, 0, 0, 0, 0, 0, 0, 1), {   )   }
    (0, 1, 0, 0, 0, 0, 0, 0, 0, 1)  { Number}
  );

Var
  Stack1 : Array [0..StackSize-1] of Char;
  Stack2 : Array [0..StackSize-1] of Real;

function power (x, y : real; var err : boolean) : real;
var
  pwr : real;
  n   : longint;
begin
  err := false;
  if (x = 0) and (y = 0) then
    begin
      err := true;
      exit
    end;
  if x > 0 then pwr := exp(y*ln(x))
  else if x = 0 then
    begin
      if y > 0 then pwr := 0
      else
        begin
          err := true;
          exit
        end
    end
  else if y <> int(y) then
    begin
      err := true;
      exit
    end
  else if y = 0 then pwr := 1
  else
    begin
      pwr := x;
      n := round(abs(y));
      while n > 1 do
        begin
          pwr := pwr*x;
          dec (n)
        end;
      if y < 0 then pwr := 1/pwr
    end;
  power := pwr
end;

Function Push1 (C : Char) : Boolean;
Begin
  If SP1 < StackSize-1 then Begin
    Stack1[SP1] := C;
    Inc (SP1);
    Push1 := True
  End Else
    Push1 := False
End;

Function Pop1 (Var C : Char) : Boolean;
Begin
  If SP1 > 0 then Begin
    Dec (SP1);
    C := Stack1[SP1];
    Pop1 := True
  End Else
    Pop1 := False
End;

Function Push2 (X : Real) : Boolean;
Begin
  If SP2 < StackSize-1 then Begin
    Stack2[SP2] := X;
    Inc (SP2);
    Push2 := True
  End Else
    Push2 := False
End;

Function Pop2 (Var X : Real) : Boolean;
Begin
  If SP2 > 0 then Begin
    Dec (SP2);
    X := Stack2[SP2];
    Pop2 := True
  End Else
    Pop2 := False
End;

Function Parse (S : String; Arg : Real; Var Result : Real) : Byte;
Var
  I, Act : Byte;
  SX : String;
  X, Y, Z : Real;
  Op : Char;
  Err : Integer;
  NumPhase,
  Success,
  EType,
  LS : Byte;
Function OpNumber (Op : Char) : Byte;
Var I : Byte;
Begin
  For I := 0 to SizeOf(Set1) do
    If Set1[I] = Op then Break;
  OpNumber := I
End;
Function DoMath (Op : Char) : Boolean;
Var
  Err : Boolean;
Begin
  Err := False;
  Case Op of
    '+' : Z := X + Y;
    '-' : Z := X - Y;
    '*' : Z := X * Y;
    '/' :
      Begin
        Err := Y = 0;
        If not Err then Z := X / Y
      End;
    '%' :
      Begin
        Err := Y = 0;
        If not Err then Z := X - Y*Int(X/Y)
      End;
    '^' : Z := Power (X, Y, Err);
  End;
  DoMath := not Err
End;
Label LRep, LEnd;
Begin
  Success := 0;                                 { OK }
  S := S + #0;
  LS := Length (S);
  SP1 := 0; SP2 := 0;
  I := 1;
  EType := 0;                                   { Empty }
  Repeat
    If S[I] in ['0'..'9'] then Begin
      If Table2[EType,9] = 1 then Begin
        Success := 4;                           { Positional error }
        Goto LEnd
      End
      Else EType := 9;
      SX := '';
      While (S[I] in ['0'..'9']) and (I < LS) do Begin
        SX := SX+S[I];
        Inc (I)
      End;
      NumPhase := 1;
      If (S[I] = '.') and (I < LS) then Begin
        SX := SX+'.';
        NumPhase := 2;
        Inc (I);
        While (S[I] in ['0'..'9']) and (I < LS) do Begin
          SX := SX+S[I];
          NumPhase := 3;
          Inc (I)
        End
      End;
      If NumPhase = 2 then Begin
        Success := 1;                           { Error in numeric format }
        Goto LEnd;
      End;
      If (S[I] in ['e','E']) and (I < LS) then Begin
        SX := SX+'E';
        NumPhase := 4;
        Inc (I);
        If (S[I] in ['-','+']) and (I < LS) then Begin
          If S[I] = '-' then SX := SX+'-';
          Inc (I)
        End;
        While (S[I] in ['0'..'9']) and (I < LS) do Begin
          SX := SX+S[I];
          NumPhase := 5;
          Inc (I)
        End;
        If NumPhase = 4 then Begin
          Success := 1;                         { Error in numeric format }
          Goto LEnd;
        End;
      End;
      Val (SX, X, Err);
      Push2 (X)
    End
    Else If S[I] in ['x','X'] then Begin
      If Table2[EType,9] = 1 then Begin
        Success := 4;                           { Positional error }
        Goto LEnd
      End
      Else EType := 9;
      Push2 (Arg);
      Inc (I)
    End
    Else If S[I] in [#0,'(','+','-','*','/','%','^',')'] then Begin
LRep:
      Act := OpNumber (S[I]);
      If ((Act = 2) or (Act = 3)) and ((EType = 0) or (EType = 1)) then
        Push2 (0);
      If Table2[EType,Act] = 1 then Begin
        Success := 4;                           { Positional error }
        Goto LEnd
      End
      Else EType := Act;
      If SP1 = 0 then Act := Table1[0,Act]
      Else Act := Table1[OpNumber(Stack1[SP1-1]),Act];
      Case Act of
        1 : Push1 (S[I]);
        3 : Pop1 (Op);
        2, 4 :
          Begin
            Pop1 (Op); Pop2 (Y); Pop2 (X);
            If not DoMath (Op) then Begin
              Success := 6;                     { Math error }
              Goto LEnd
            End;
            Push2 (Z);
            If Act = 2 then Push1 (S[I])
            Else Begin
              EType := 9;
              Goto LRep
            End
          End;
        5 :
          Begin
            Success := 3;                       { Not balanced brakets }
            Goto LEnd
          End
      End;
      Inc (I)
    End
    Else If S[I] in [#32, #9] then
      Inc (I)
    Else Begin
      Success := 2;                             { Unknown symbol }
      Goto LEnd
    End
  Until I > LS;
  If not Pop2 (Result) then
    Success := 5;                               { Expression is empty }
LEnd:
  Parse := Success
End;

End.
